import data
import pygame, sys, os, random

from pygame.locals import *

FPS = 30
SCREEN_MODE = (1024,600)
BACKGROUND_COLOR = (0,0,0)
TERRAIN_COLOR = (255,0,0)
WATER_COLOR = (0,0,255, 25)
SELECTED_SCROLL_COLOR = (0,255,0)
COLUMN_HEIGHT = 150
RAISING_SPEED = 10
GAME_TITLE = "Terraformancer Version 0.7 by Trobadour for PyWeek 10"
PEOPLE_SIZE = (17,26) # width, height
PEOPLE_SPEED = 1
TOWER_SIZE = (30,112) # width, height
PRINCESS_SIZE = (17,39) # width, height
TASK_SYMBOL_SIZE = (17,17)
SPELLBAR_TOPLEFT = (150,500)
SPELLBAR_SCROLL_SIZE = (80,80)

GFX_EARTHQUAKE = "earthquake.png"
GFX_BACKGROUND = "background.png"
GFX_GAME_OVER = "princess_rescued.png"
GFX_PAUSED = "pause.png"
GFX_TIMEOUT = "timeover.png"
GFX_HELP = "help.png"
GFX_TERRAIN_01 = "terrain_mud.png"
GFX_PEOPLE_01 = "people_right_side.png"
GFX_PEOPLE_02 = "people_front.png"
GFX_TOWER = "tower.png"
GFX_SCROLL_UP = "scroll_up.png"
GFX_SCROLL_DOWN = "scroll_down.png"
GFX_SCROLL_TO_PRINCESS = "scroll_to_princess.png"
GFX_SCROLL_PAUSE = "scroll_pause.png"
GFX_SCROLL_HELP = "scroll_help.png"
GFX_SCROLL_RESTART = "scroll_restart.png"
GFX_SCROLL_QUIT = "scroll_quit.png"
GFX_PRINCESS = "princess.png"
GFX_TASK_SYMBOL_WAIT = "task_wait.png"
GFX_TASK_SYMBOL_PRINCESS = "task_princess.png"
GFX_TREE = "tree.png"

SFX_MUSIC_02 = "Pedro Collares (Organic Healing Sounds) - 03 - Gothic.ogg"
SFX_EARTHQUAKE = "earthquake.ogg"

SCROLL_RAISE_KEY = "Cast Raise"
SCROLL_LOWER_KEY = "Cast Lower"
SCROLL_TO_PRINCESS_KEY = "Goto Princess"
SCROLL_PAUSE_KEY = "Pause"
SCROLL_HELP_KEY = "HELP"
SCROLL_RESTART_KEY = "Restart"
SCROLL_QUIT_KEY = "Quit"

PEOPLE_TASK_WAIT = 0
PEOPLE_TASK_MOVE_LEFT = 1
PEOPLE_TASK_MOVE_RIGHT = 2
NOF_PEOPLE_TASKS = 3

COLUMN_TASK_NONE = 0
COLUMN_TASK_RAISE = 1
COLUMN_TASK_LOWER = 2

PEOPLE_TYPE_PERSON = "Person"
PEOPLE_TYPE_GUY = "Guy"
PEOPLE_TYPE_PRINCESS = "Princess"

# Global functions
def load_image( file, new_width=0, new_height=0 ):
    # create filename
    filename = os.path.join( 'data', file )
    # load the image
    try:
        surface = pygame.image.load( filename )
    except pygame.error:
        raise SystemExit, 'Could not load image "%s" %s'%( filename, pygame.get_error())
    # convert and resize
    image = surface.convert_alpha()
    if new_width>0 and new_height>0:
        image = pygame.transform.scale( image, (new_width,new_height) )
    return image
  
def load_sound( file ):
  # create a filename
  filename = os.path.join( 'data', file )
  
  # load the sound
  try:
    return pygame.mixer.Sound( filename )
  except pygame.error:
    raise SystemExit, 'Could not load sound "%s" %s'%( filename, pygame.get_error())

class SpellBar:
    def __init__(self, rect ):
        self.rect = rect
        self.keys = []
        self.current_spell = SCROLL_RAISE_KEY
        self.textures = {}
        self.texture_rect = rect.copy()
        self.texture_rect.width = self.texture_rect.height
        self.spacing = 30
        return
    
    def add_button(self, key, texture_file):
        self.keys.append( key )
        self.textures[key] = load_image(texture_file, self.texture_rect.width, self.texture_rect.height)
        return
    
    def paint(self,screen):
        for key in self.keys:
            screen.blit( self.textures[key], self.scroll_rect(key) )
            if key is self.current_spell:
                pygame.draw.rect( screen, SELECTED_SCROLL_COLOR, self.scroll_rect(key), 2 )
        return
    
    def scroll_rect(self,key):
        rect = pygame.Rect(0,0,0,0)
        index = self.keys.index(key)
        if index is None:
            return rect
        rect = self.texture_rect.copy()
        if index is 0:
            rect.left = self.rect.left
        else:
            rect.left = self.rect.left + index*(rect.width+self.spacing)
        return rect
    
    def handle_event(self,event):
        if event.type is MOUSEBUTTONUP and event.button is 1:
            for key in self.keys:
                rect = self.scroll_rect(key)
                if rect.collidepoint(event.pos):
                    self.current_spell = key
                    return True
        return False

class Resources:
    def __init__(self):
        self.columns = []
        self.people = []
        self.earthquake_sfx = load_sound(SFX_EARTHQUAKE)
    
    def add_column(self,column):
        self.columns.append( column )
    
    def nof_columns(self):
        return self.columns.__len__()
    
    def standing_on(self,guy):
        for column in self.columns:
            if guy.rect.centerx >= column.rect.left and guy.rect.centerx <= column.rect.right:
                return column
        return None
    
    def nof_guys(self):
        return self.people.__len__()
    
    def tower( self ):
        for column in self.columns:
            if column.tower is True:
                return column
        return None
    
    def princess( self ):
        for guy in self.people:
            if guy.type is PEOPLE_TYPE_PRINCESS:
                return guy
        return None

class Person:
    def __init__(self, rect, resources):
        self.resources = resources
        
        self.MaxClimableHeightDiff = 20
        self.MaxWaterDepth = 20
       
        self.rect = rect
        self.textures = {}
        
        self.textures[GFX_TASK_SYMBOL_WAIT] = load_image( GFX_TASK_SYMBOL_WAIT, TASK_SYMBOL_SIZE[0], TASK_SYMBOL_SIZE[1] )

        self.task = PEOPLE_TASK_WAIT
        self.task_ends_in = -1
        
        self.type = PEOPLE_TYPE_PERSON
        
        return
    
    def handle_event( self, event, spell ):
        return False
    
    def begin_new_task(self):
        self.task = random.randint(0,NOF_PEOPLE_TASKS-1)
        self.task_ends_in = random.randint( 2*FPS, 8*FPS )
        return
    
    def do_wait(self):
        self.task = PEOPLE_TASK_WAIT
        self.task_ends_in = random.randint( 2*FPS, 8*FPS )
        return
    
    def paint(self, screen):
        screen.blit( self.textures[self.task], self.rect )
        self.paint_task_symbol( screen )
        return
    
    def paint_task_symbol( self, screen ):
        symbol_rect = self.textures[self.task].get_rect()
        symbol_rect.bottom = self.rect.top
        symbol_rect.centerx = self.rect.centerx
            
        if self.task is PEOPLE_TASK_WAIT:
            screen.blit( self.textures[GFX_TASK_SYMBOL_WAIT], symbol_rect )
        return
    
    def can_move_left(self):
        column = self.resources.standing_on(self)
        if column is None:
            return False
        if column.rect.left < self.rect.centerx:
            return True
        else:
            next_column = column.column_to_the_left()
            if next_column is None:
                return False
            if next_column.water > self.MaxWaterDepth:
                return None
            if abs(column.height_diff(next_column)) > self.MaxClimableHeightDiff:
                return False
            else:
                return True
    
    def can_move_right(self):
        column = self.resources.standing_on(self)
        if column is None:
            return False
        if column.rect.right > self.rect.centerx:
            return True
        else:
            next_column = column.column_to_the_right()
            if next_column is None:
                return False
            if next_column.water > self.MaxWaterDepth:
                return None
            if abs(column.height_diff(next_column)) > self.MaxClimableHeightDiff:
                return False
            else:
                return True
    
    def do_task(self):
        if self.task_ends_in > 0:
            self.task_ends_in = self.task_ends_in - 1
        if self.task is PEOPLE_TASK_WAIT:
            return
        if self.task is PEOPLE_TASK_MOVE_LEFT:
            if self.can_move_left():
                self.rect = self.rect.move( -PEOPLE_SPEED, 0 )
            else:
                self.do_wait()
            return
        if self.task is PEOPLE_TASK_MOVE_RIGHT:
            if self.can_move_right():
                self.rect = self.rect.move( PEOPLE_SPEED, 0 )
            else:
                self.do_wait()
            return
        return
        
    def update(self, column):
        if self.task_ends_in == 0:
            self.begin_new_task()
        else:
            self.do_task()
        
        if column is None:
            return
        self.rect.bottom = column.rect.top
        return

class Guy( Person ):
    def __init__( self, rect, resources ):
        Person.__init__(self,rect,resources)
        
        self.textures[PEOPLE_TASK_WAIT] = load_image( GFX_PEOPLE_02, rect.width, rect.height )
        self.textures[PEOPLE_TASK_MOVE_LEFT] = pygame.transform.flip( load_image(GFX_PEOPLE_01,rect.width,rect.height), True, False )
        self.textures[PEOPLE_TASK_MOVE_RIGHT] = load_image( GFX_PEOPLE_01, rect.width, rect.height )
        self.textures[GFX_TASK_SYMBOL_PRINCESS] = load_image( GFX_TASK_SYMBOL_PRINCESS, TASK_SYMBOL_SIZE[0], TASK_SYMBOL_SIZE[1] )
        
        self.find_princess = False
        self.deliver_princess = False
        self.task_ends_in = random.randint( 0, 2*FPS )
        
        self.type = PEOPLE_TYPE_GUY

        return
    
    def paint( self, screen ):
        Person.paint( self, screen )

        if self.find_princess is True or self.deliver_princess is True:
            symbol_rect = self.textures[self.task].get_rect()
            symbol_rect.bottom = self.rect.top
            symbol_rect.centerx = self.rect.centerx
            screen.blit( self.textures[GFX_TASK_SYMBOL_PRINCESS], symbol_rect )
        return
    
    def handle_event( self, event, spell ):
        if event.type is not MOUSEBUTTONDOWN or event.button is not 1:
            return False
        if self.rect.collidepoint(event.pos):
            if spell is SCROLL_TO_PRINCESS_KEY:
                self.find_princess = not self.find_princess
                
                if self.find_princess is not True:
                    self.begin_new_task()
                else:
                    self.task_ends_in = 0
                return True
        return False
    
    def update( self, column ):
        if column is not None:
                self.rect.bottom = column.rect.top
        if self.deliver_princess is True and self.task_ends_in is 0:
            tower = self.resources.tower()
            if tower.rect.centerx < self.rect.centerx:
                self.task = PEOPLE_TASK_MOVE_LEFT
                self.task_ends_in = random.randint( 2*FPS, 8*FPS )
            else:
                self.task = PEOPLE_TASK_MOVE_RIGHT
                self.task_ends_in = random.randint( 2*FPS, 8*FPS )
            return
        
        if self.find_princess is True:
            princess = self.resources.princess()
            if princess is None:
                return
            
            if abs(princess.rect.centerx-self.rect.centerx) < 2:
                self.find_princess = False
                self.deliver_princess = True
                self.task_ends_in = 0
                princess.follow_guy( self )
            elif princess.rect.centerx > self.rect.centerx:
                self.task = PEOPLE_TASK_MOVE_RIGHT
                self.task_ends_in = -1
            else:
                self.task = PEOPLE_TASK_MOVE_LEFT
                self.task_ends_in = -1
        Person.update( self, column )
        return

class Princess( Person ):
    def __init__(self, rect, resources):
        Person.__init__(self,rect,resources)
        self.textures[PEOPLE_TASK_WAIT] = load_image( GFX_PRINCESS, PRINCESS_SIZE[0],PRINCESS_SIZE[1] )
        self.textures[PEOPLE_TASK_MOVE_LEFT] = load_image( GFX_PRINCESS, PRINCESS_SIZE[0],PRINCESS_SIZE[1] )
        self.textures[PEOPLE_TASK_MOVE_RIGHT] = load_image( GFX_PRINCESS, PRINCESS_SIZE[0],PRINCESS_SIZE[1] )
        
        self.arrived = False
        self.task_ends_in = -1
        self.follow = None
        self.type = PEOPLE_TYPE_PRINCESS
    
    def follow_guy( self, guy ):
        self.follow = guy
        self.task_ends_in = 0
    
    def update( self, column ):
        tower = self.resources.tower()
        if abs(tower.rect.centerx-self.rect.centerx)<2:
            self.arrived = True
            return
    
        if self.follow is not None and abs(self.follow.rect.centerx-self.rect.centerx)<20:
            if self.task_ends_in is 0:
                tower = self.resources.tower()
                if tower.rect.centerx < self.rect.centerx:
                    self.task = PEOPLE_TASK_MOVE_LEFT
                    self.task_ends_in = random.randint( 2*FPS, 8*FPS )
                else:
                    self.task = PEOPLE_TASK_MOVE_RIGHT
                    self.task_ends_in = random.randint( 2*FPS, 8*FPS )
        else:
            self.task = PEOPLE_TASK_WAIT
            self.task_ends_in = random.randint( 2*FPS, 8*FPS )
        
        Person.update( self, column )
    
class Column:
    def __init__(self, rect, resources):
        self.rect = rect
        self.water = 0
        self.tree = False
        self.resources = resources
        
        self.task = COLUMN_TASK_NONE
        self.task_ends_in = 0
        
        self.tower = False
        
        self.ColumnMaxHeightDiff = 40
        
        # load texture
        self.texture = load_image( GFX_TERRAIN_01, rect.width, rect.height )
        
        self.tower_texture = load_image( GFX_TOWER,TOWER_SIZE[0], TOWER_SIZE[1] )
        self.tree_texture = load_image( GFX_TREE, 60, 60 )
        
        return
    
    def paint(self, screen):
        # draw terrain
        screen.blit( self.texture, self.rect )
        
        # draw tower
        if self.tower is True:
            rect = pygame.Rect((0,0),TOWER_SIZE)
            rect.bottom = self.rect.top
            rect.centerx = self.rect.centerx
            screen.blit( self.tower_texture, rect )
        
        if self.tree is True:
            rect = pygame.Rect((0,0),(60,60))
            rect.bottom = self.rect.top
            rect.centerx = self.rect.centerx
            screen.blit( self.tree_texture, rect )
        
        # draw water
        if self.water > 0:
            water_rect = self.rect.copy()
            water_rect.height = self.water
            water_rect.bottomleft = self.rect.topleft
            water_rect.width = water_rect.width + 1
            #screen.fill( WATER_COLOR, water_rect );
            pygame.draw.rect( screen, WATER_COLOR, water_rect )
        return
    
    def update(self):
        if self.task_ends_in > 0:
            self.task_ends_in = self.task_ends_in - 1
            if self.task is COLUMN_TASK_RAISE:
                self.rect = self.rect.move( 0, -1 )
            if self.task is COLUMN_TASK_LOWER:
                self.rect = self.rect.move( 0, 1 )
        else:
            self.task = COLUMN_TASK_NONE
        
        self.update_water()
        return
    
    def update_water(self):
        
        if self.water <= 0:
            return
        
        left_neighbour = None
        right_neighbour = None
        
        # determine neighbours
        index = self.resources.columns.index(self)
        if index > 0:
            left_neighbour = self.resources.columns[index-1]
        if index < self.resources.columns.__len__()-1:
            right_neighbour = self.resources.columns[index+1]
            
        
        if left_neighbour is not None:
            my_water_surface_y = self.rect.top-self.water
            # water surface position of left neighbour
            left_water_surface_y = left_neighbour.rect.top-left_neighbour.water
            # let water flow to the left if possible
            if left_water_surface_y > my_water_surface_y:
                left_neighbour.water = left_neighbour.water + 1
                self.water = self.water - 1
        if self.water <= 0:
            return
        if right_neighbour is not None:
            my_water_surface_y = self.rect.top-self.water
            # water surface position of right neighbour
            right_water_surface_y = right_neighbour.rect.top-right_neighbour.water
            # let water flow to the right if possible
            if right_water_surface_y > my_water_surface_y:
                right_neighbour.water = right_neighbour.water + 1
                self.water = self.water - 1
            
        return
    
    def raise_up(self):
        if self.task is COLUMN_TASK_NONE:
            self.task = COLUMN_TASK_RAISE
            self.task_ends_in = RAISING_SPEED
        return
    
    def lower_down(self):
        if self.task is COLUMN_TASK_NONE:
            self.task = COLUMN_TASK_LOWER
            self.task_ends_in = RAISING_SPEED
        return
    
    def column_to_the_left(self):
        if self.resources.columns.count(self) is 0:
            return None
        index = self.resources.columns.index(self)
        if index is 0:
            return None
        else:
            return self.resources.columns[index-1]
    
    def column_to_the_right(self):
        if self.resources.columns.count(self) is 0:
            return None
        index = self.resources.columns.index(self)
        if index is self.resources.nof_columns()-1:
            return None
        else:
            return self.resources.columns[index+1]
        
    def height_diff(self,neighbour_column):
        if neighbour_column is None:
            return 2*self.ColumnMaxHeightDiff
        
        my_height = self.rect.top
        neighbour_column_height = neighbour_column.rect.top
        
        return my_height-neighbour_column_height
    
    def handle_event(self, event, spell):
        if spell is not SCROLL_RAISE_KEY and spell is not SCROLL_LOWER_KEY:
            return False
        if event.type is MOUSEBUTTONDOWN and event.button is 1:
            if self.rect.collidepoint(event.pos):
                if spell is SCROLL_RAISE_KEY:
                    self.raise_up()
                elif spell is SCROLL_LOWER_KEY:
                    self.lower_down()
                return True
        return False

class LevelData:
    def __init__(self):
        self.name = ""
        self.nof_columns = 0
        self.column_width = []
        self.column_surface_position = []
        self.column_water = []
        self.nof_guys = 0
        self.people_start_column = []
        self.tower_at = 0
        self.princess_start_column = 0
        self.trees = []
        self.time_limit = 0
        self.earthquakes = []

class Level:
    """Contains the """
    def __init__(self, leveldata):
        self.resources = Resources()
        self.data = leveldata
        self.init_level( self.data )

    def init_level(self, leveldata):
        self.spellbar = SpellBar( pygame.Rect(SPELLBAR_TOPLEFT,SPELLBAR_SCROLL_SIZE) )
        self.spellbar.add_button( SCROLL_RAISE_KEY, GFX_SCROLL_UP )
        self.spellbar.add_button( SCROLL_LOWER_KEY, GFX_SCROLL_DOWN )
        self.spellbar.add_button( SCROLL_TO_PRINCESS_KEY, GFX_SCROLL_TO_PRINCESS )
        self.spellbar.add_button( SCROLL_PAUSE_KEY, GFX_SCROLL_PAUSE )
        self.spellbar.add_button( SCROLL_HELP_KEY, GFX_SCROLL_HELP )
        self.spellbar.add_button( SCROLL_RESTART_KEY, GFX_SCROLL_RESTART )
        self.spellbar.add_button( SCROLL_QUIT_KEY, GFX_SCROLL_QUIT )
        
        # determine width of terrain
        self.terrain_width = 0
        for i in range(self.data.nof_columns):
            self.terrain_width = self.terrain_width + self.data.column_width[i]
        
        # determine border
        border = SCREEN_MODE[0] - (self.terrain_width+self.data.nof_columns-1)
        
        # determine the terrain boundaries
        self.terrain_left = border/2
        self.terrain_right = self.terrain_left + self.terrain_width
        
        # create columns
        current_width = 0
        for i in range(self.data.nof_columns):
            # create a rect to describe the columns position
            left = self.terrain_left+current_width
            top = self.data.column_surface_position[i]
            width = self.data.column_width[i]
            height = COLUMN_HEIGHT
            rect = pygame.Rect( left, top, width, height )
            
            # create column
            column = Column(rect,self.resources)
            column.water = self.data.column_water[i]
            
            if self.data.trees.count(i) > 0:
                column.tree = True
            
            self.resources.columns.append( column )
            
            current_width = current_width + self.data.column_width[i] + 1
            
        # place tower
        self.resources.columns[self.data.tower_at].tower = True
        
        # create and place people
        people = []
        for i in range(self.data.nof_guys):
            # create a rect for the guy
            column = self.resources.columns[ self.data.people_start_column[i] ]
            
            rect = pygame.Rect(1,1,PEOPLE_SIZE[0],PEOPLE_SIZE[1])
            rect.bottom = column.rect.top
            rect.centerx = random.randint(column.rect.left,column.rect.right)
            
            self.resources.people.append( Guy(rect,self.resources) )
            
        # place princess
        column = self.resources.columns[ self.data.princess_start_column ]
        princess_rect = pygame.Rect(1,1,PRINCESS_SIZE[0],PRINCESS_SIZE[1])
        princess_rect.bottom = column.rect.top
        princess_rect.centerx = random.randint(column.rect.left,column.rect.right)
        self.resources.people.append( Princess(princess_rect,self.resources) )
        
        self.current_time = self.data.time_limit
        
        return
    
    def paint( self, screen ):
        self.spellbar.paint( screen )

        if self.spellbar.current_spell is not SCROLL_HELP_KEY:
            # draw terrain columns
            for column in self.resources.columns:
                column.paint( screen )
            for guy in self.resources.people:
                guy.paint( screen )
        
            # paint timer
            left = 10
            right = 1014
            width = 1000
            top = 450
            mid = 460
            bottom = 470
        
            progress = width * float(self.current_time)/float(self.data.time_limit)
            
            pygame.draw.line( screen, (255,0,0), (left,top), (left,bottom), 2 )
            pygame.draw.line( screen, (255,0,0), (right,top), (right,bottom), 2 )
            pygame.draw.line( screen, (255,0,0), (left,mid), (left+progress,mid) )
            for earthquake in self.data.earthquakes:
                progress = width * float(earthquake)/float(self.data.time_limit)
                pygame.draw.line( screen, (255,0,0), (left+progress,top), (left+progress,bottom), 2 )
        return
    
    def handle_event(self, event):
        current_spell = self.spellbar.current_spell
        for column in self.resources.columns:
            if column.handle_event(event,current_spell) is True:
                return True
        if self.spellbar.handle_event(event) is True:
            return True
        for guy in self.resources.people:
            if guy.handle_event(event,self.spellbar.current_spell) is True:
                return True
        return False
    
    def update(self):
        # update columns
        for column in self.resources.columns:
            column.update()
        
        if self.data.earthquakes.count(self.current_time)>0:
            self.resources.earthquake_sfx.play()
            for column in self.resources.columns:
                for i in range(8):
                    if random.randint(0,1) is 0:
                        column.raise_up()
                    else:
                        column.lower_down()

        # update people
        for guy in self.resources.people:
            column = self.resources.standing_on( guy )
            guy.update( column )
        
        self.current_time = self.current_time - 1
        pass

def main():
    pygame.mixer.pre_init(22050, 16, 2, 1024)
          
    pygame.init()
    pygame.mixer.init()
    window = pygame.display.set_mode( SCREEN_MODE )
    pygame.display.set_caption( GAME_TITLE )
    screen = pygame.display.get_surface()
    
    file = os.path.join('data',SFX_MUSIC_02)
    pygame.mixer.music.load( file )
    pygame.mixer.music.play( -1 )
    
    # Create data describing level 1
    level1 = LevelData()
    level1.name = "Level 1"
    level1.nof_columns = 15
    level1.tower_at = 0

    level1.column_width = [60,60,60,60,60,60,60,60,60,60,60,60,60,60,60]
    level1.column_surface_position = [250,240,290,300,260,250,250,260,240,320,270,220,250,240,260]
    level1.column_water = [0,0,30,40,0,0,0,0,0,50,0,0,0,0,0]

    level1.nof_guys = 1
    level1.people_start_column.append( 1 )
    level1.princess_start_column = 14
    level1.trees.append( 1 )
    level1.trees.append( 4 )
    level1.trees.append( 7 )
    level1.trees.append( 8 )
    level1.trees.append( 12 )
    level1.time_limit = 90 * FPS
    level1.earthquakes = [10*FPS,20*FPS,30*FPS,40*FPS,50*FPS,60*FPS,70*FPS,80*FPS]

    # Create Level 1
    level = Level( level1 )
    level.spellbar.current_spell = SCROLL_PAUSE_KEY
    
    background = load_image( GFX_BACKGROUND, SCREEN_MODE[0], SCREEN_MODE[1]  )
    rescued_screen = load_image( GFX_GAME_OVER, SCREEN_MODE[0], SCREEN_MODE[1]  )
    paused_screen = load_image( GFX_PAUSED, SCREEN_MODE[0], SCREEN_MODE[1]  )
    help_screen = load_image( GFX_HELP, SCREEN_MODE[0], SCREEN_MODE[1]  )
    timeout_screen = load_image( GFX_TIMEOUT, SCREEN_MODE[0], SCREEN_MODE[1]  )
    
    run_game = True
    finished = False
    timeout = False
    
    while run_game:
        ticks_start = pygame.time.get_ticks()
        
        if finished is False:
            finished = level.resources.princess().arrived
        
        screen.fill( BACKGROUND_COLOR )
        
        if finished is True:
            screen.blit( rescued_screen, (0,0) )
        elif timeout is True:
            screen.blit( timeout_screen, (0,0) )
        elif level.spellbar.current_spell is SCROLL_PAUSE_KEY:
            screen.blit( paused_screen, (0,0) )
        elif level.spellbar.current_spell is SCROLL_HELP_KEY:
            screen.blit( help_screen, (0,0) )
        else:
            screen.blit( background, (0,0) )
        
        # Handle events
        for event in pygame.event.get():
            if level.handle_event(event) is True:
                continue
            if event.type is QUIT:
                run_game = False

        if level.spellbar.current_spell is SCROLL_RESTART_KEY:
            level = Level( level1 )
            finished = False
            timeout = False
        
        if level.spellbar.current_spell is not SCROLL_PAUSE_KEY and level.spellbar.current_spell is not SCROLL_HELP_KEY and timeout is False and finished is False:
            # Update level objects
            level.update()

        if level.current_time <= 0:
            timeout = True

        if level.spellbar.current_spell is SCROLL_QUIT_KEY:
            run_game = False
        
        # Paint level and all objects
        level.paint( screen )
        
        pygame.display.flip()
        
        # wait a little bit (if required) to have 30 fps
        ticks_end = pygame.time.get_ticks()
        ticks_diff = ticks_end - ticks_start
        if ticks_diff < 1000/FPS:
            pygame.time.delay( 1000/FPS-ticks_diff )
    
